//
//  acl.cpp  --  part of LNFC library
//
//  Implementation of classes
//    Database_ACL, Database_ACL_d, ACL_entry, ACL_entry_d
//
//  Written by Evseev at Jul 2001
//

#include "acl.h"

#include "_lnfc.h"

#include <acl.h>
#include <nsfdata.h>  // DBHANDLE
#include <nsfdb.h>    // NSFDbReadACL, NSFDbUpdateACL
#include <nsferr.h>   // ERR_HDL_NULL


#define RETURN_IF_ACL_ISNOT_READY() \
        do {                        \
            RETURN_IF_BAD(acl);     \
            if (!acl->is_open())    \
                RETURN_API_ERROR(ERR_HDL_NULL); \
        } while(0)


//---------------------------------------------------------------------

// ACL_entry_d implementation


ACL_entry_d::ACL_entry_d(
    Database_ACL               &acl_arg,
    const Notes_name           &name,
    access_level_t             level,
    const access_privileges_t &privileges,
    access_flags_t             flags,
    bool                       must_be_created)
  :
    acl              (acl_arg),
    entry_name       (name),
    access_level     (level),    
  //access_privileges(privileges),
    access_flags     (flags),    
    what_to_update   (0)
{
    memcpy(access_privileges, privileges, sizeof(access_privileges));

    if (must_be_created)
        create();
}


int ACL_entry_d::create(bool)
{
    RETURN_IF_ACL_ISNOT_READY();

    reset();
    
    RETURN_IF_API_ERROR(
        ACLAddEntry(
            (HANDLE)acl->get_handle(),
            entry_name->plain(),
            access_level,
            (ACL_PRIVILEGES*)&access_privileges,
            access_flags
    )   );

    acl->must_be_saved();

    return 0;
}


int ACL_entry_d::remove(bool)
{
    RETURN_IF_ACL_ISNOT_READY();

    RETURN_IF_API_ERROR(
        ACLDeleteEntry(
            (HANDLE)acl->get_handle(),
            entry_name->plain()
    )   );

    return 0;
}


//---------------------------------------------------------------------

// Database_ACL_d members


Database_ACL_d::Database_ACL_d(const Database &db_arg, bool ignore_old_acl_arg)
  : db(db_arg),
    ignore_old_acl(ignore_old_acl_arg)
{
    dont_save();

    if (ignore_old_acl) {
        create();
    } else {
        open();
    }
}


int Database_ACL_d::open()
{
    RETURN_IF_DB_ISNOT_READY();

    reset();

    HANDLE htmp;
    RETURN_IF_API_ERROR(NSFDbReadACL( (DBHANDLE) db->get_handle(), &htmp));
    
    handle = (notes_handle_t)htmp;
    return 0;
}


int Database_ACL_d::create()
{
    reset();

    HANDLE htmp;
    RETURN_IF_API_ERROR(ACLCreate(&htmp));

    handle = (notes_handle_t)htmp;
    must_be_saved();
    return 0;
}


int Database_ACL_d::save(bool)
{
    RETURN_IF_DB_ISNOT_READY();
    RETURN_IF_ERROR();
    RETURN_IF_NULL_HANDLE();

    HANDLE htmp;
    RETURN_IF_API_ERROR(
        NSFDbStoreACL(
            (DBHANDLE) db->get_handle(),
            (HANDLE) handle,
            0L,  // ObjectID, reserved, must be 0L
            ignore_old_acl));
    
    return 0;
}


//---------------------------------------------------------------------

// Database_ACL_d enumeration members


typedef struct {
    int (*action)(const ACL_entry &e, void *args);
    void *action_args;
    Database_ACL *acl_addr;
} Foreach_args;


static void LNPUBLIC foreach_callback(
    void *args_addr,
    char *entry_name,
    WORD  access_level,
    ACL_PRIVILEGES *privileges,
    WORD  access_flags)
{
    Foreach_args *a = reinterpret_cast<Foreach_args*>(args_addr);

    access_privileges_t privs;
    memcpy(privs, privileges, sizeof(privs));

    ACL_entry acle(*a->acl_addr,
                   nSTR entry_name,
                   (access_level_t)access_level,
                   privs,
                   (access_flags_t)access_flags); 

    a->action(acle, a->action_args);
}


int Database_ACL_d::foreach_entry(int (*action)(const ACL_entry &e, void *args), void *args)
{
    RETURN_IF_DB_ISNOT_READY();
    RETURN_IF_ERROR();
    RETURN_IF_NULL_HANDLE();

    Database_ACL acl(this);

    Foreach_args a;
    a.action = action;
    a.action_args = args;
    a.acl_addr = &acl;

    RETURN_IF_API_ERROR(
        ACLEnumEntries(
            (HANDLE)handle,
            foreach_callback,
            &a
    )   );

    return 0;
}


//---------------------------------------------------------------------

// Database_ACL_d enumeration members


typedef enum { Match_name, Match_level } Match_t;

typedef struct {
    int (*action)(const ACL_entry &e, void *args);
    void *action_args;
    Database_ACL *acl_addr;
    const Notes_name *name_addr;
    access_level_t level;
    Match_t match;
} Foreach_matching_args;


static void LNPUBLIC foreach_matching_callback(
    void *args_addr,
    char *entry_name,
    WORD  access_level,
    ACL_PRIVILEGES *privileges,
    WORD  access_flags)
{
    Foreach_matching_args *a = reinterpret_cast<Foreach_matching_args*>(args_addr);
    Notes_string ename(entry_name);
    
    if ( ( (a->match == Match_name) &&
           (ename == (*a->name_addr)->canonical())
         ) ||
         ( (a->match == Match_level) &&
           (a->level == (access_level_t)access_level)
       ) )
    {
        access_privileges_t privs;
        memcpy(privs, privileges, sizeof(privs));

        ACL_entry acle(*a->acl_addr,
                       ename,
                       (access_level_t)access_level,
                       privs,
                       (access_flags_t)access_flags); 

        a->action(acle, a->action_args);
    }
}


int Database_ACL_d::foreach_matching(
    const Notes_name &matching_name,
    int (*action)(const ACL_entry &e, void *args),
    void *args)
{
    RETURN_IF_ERROR();
    RETURN_IF_NULL_HANDLE();

    Database_ACL acl(this);

    Foreach_matching_args a;
    a.action = action;
    a.action_args = args;
    a.acl_addr = &acl;
    a.name_addr = &matching_name;
    a.match = Match_name;

    RETURN_IF_API_ERROR(
        ACLEnumEntries(
            (HANDLE)handle,
            foreach_matching_callback,
            &a
    )   );

    return 0;
}


int Database_ACL_d::foreach_matching(
    access_level_t matching_level,
    int (*action)(const ACL_entry &e, void *args),
    void *args)
{
    RETURN_IF_ERROR();
    RETURN_IF_NULL_HANDLE();

    Database_ACL acl(this);

    Foreach_matching_args a;
    a.action = action;
    a.action_args = args;
    a.acl_addr = &acl;
    a.level = matching_level;
    a.match = Match_level;

    RETURN_IF_API_ERROR(
        ACLEnumEntries(
            (HANDLE)handle,
            foreach_matching_callback,
            &a
    )   );

    return 0;
}


static int contains_callback(const ACL_entry&, void *arg)
{
    *reinterpret_cast<bool*>(arg) = true;
    return 0;
}


bool Database_ACL_d::contains(const Notes_name &entry_name)
{
    bool result = false;
    foreach_matching(entry_name, contains_callback, &result);
    return result;
}


// EOF //
