//
//  acl.h  --  part of LNFC library
//
//  Declaration of classes
//    Database_ACL_d, Database_ACL, ACL_entry_d, ACL_entry
//
//  Written by Evseev at Jul 2001
//

#ifndef __LNFC_ACL_H
#define __LNFC_ACL_H


#ifndef __LNFC_DATABASE_H
#include "database.h"
#endif


__LNFC_NAMESPACE_BEGIN__


// forward references //
class Database_ACL;
class Database_ACL_d;
class ACL_entry;
class ACL_entry_d;


// readable aliases for values of bool args //
const bool Ignore_old_ACL = true;


typedef enum {
    No_access = 0, Depositor, Reader, Author, Editor, Designer, Manager
} access_level_t;

typedef unsigned short  access_flags_t;

typedef char access_privileges_t[10];

const access_privileges_t Empty_privileges = { 0 };

// this is instead of separate constants
const struct {
    unsigned Disabled_creation:1;
    unsigned Is_server:1;
    unsigned Disabled_deletion:1;
    unsigned Enabled_personal_agents:1;
    unsigned Enabled_personal_folders:1;
    unsigned Is_person:1;
    unsigned Is_group:1;
    unsigned Enabled_public_folders:1;
    unsigned Enabled_scripts:1;
    unsigned Public_reader:1;
    unsigned Public_writer:1; 
    unsigned _unused1:3;       // values 0x800, 0x1000, 0x2000 are unused
    unsigned Admin_modifier:1; // 0x4000 = admin server can modify reader/author fields
    unsigned Admin_server:1;
} Access_flags = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 7, 1, 1 }; 


//---------------------------------------------------------------------

class Database_ACL_d : public Notes_object_d
{

protected:

    friend Database_ACL;

    Database_ACL_d(const Database &db, bool ignore_old_acl = false);

    virtual ~Database_ACL_d() { cleanup(); }

public:

    virtual int open();

    virtual int create();

    virtual int save(bool force = false);

    int foreach_entry(   int (*action)(const ACL_entry &e, void *args),
                         void *args = 0);

    int foreach_matching(const Notes_name &name_to_process,
                         int (*action)(const ACL_entry &e, void *args),
                         void *args = 0);

    int foreach_matching(access_level_t level_to_process,
                         int (*action)(const ACL_entry &e, void *args),
                         void *args = 0);

    bool contains(const Notes_name &entry_name);

protected:

    Database db;

    bool ignore_old_acl;

};


//---------------------------------------------------------------------


class Database_ACL : public Smart_pointer<Database_ACL_d>
{

public:

    Database_ACL(const Database &db, bool ignore_old_acl = false)
      : SmPtr(new Database_ACL_d(db, ignore_old_acl))
    { }

protected:

    friend Database_ACL_d;

    // used by Database_ACL_d::foreach_entry()
    Database_ACL(Database_ACL_d *acl_addr)
      : SmPtr(acl_addr)
    { }

};


//---------------------------------------------------------------------


class ACL_entry_d : public Notes_object_d
{

protected:

    friend ACL_entry;

    ACL_entry_d(Database_ACL &acl,
                const Notes_name          &entry_name,
                access_level_t             level,
                const access_privileges_t &privileges,
                access_flags_t             flags,
                bool                       must_be_created = false);

public:

    virtual int create(bool force = false);

    virtual int remove(bool force = false);

    virtual bool exist() { return acl->contains(entry_name); }

public:

    Notes_string   get_name()  const { return entry_name;   }
    access_level_t get_level() const { return access_level; }
    access_flags_t get_flags() const { return access_flags; }

private:

    Database_ACL acl;
    Notes_name           entry_name;
    access_level_t       access_level;
    access_privileges_t  access_privileges;
    access_flags_t       access_flags;

    int what_to_update;

};


//---------------------------------------------------------------------


class ACL_entry : public Smart_pointer<ACL_entry_d>
{

public:

    ACL_entry(Database_ACL &acl,
              const Notes_name &entry_name,
              access_level_t    level = No_access,
              access_flags_t    flags = 0)
      : SmPtr(new ACL_entry_d(acl, entry_name, level, Empty_privileges, flags))
    { }

    
    // this is public for visibility from foreach_callback().
    // don't call it from other places.
    ACL_entry(Database_ACL &acl,
              const Notes_name          &entry_name,
              access_level_t             level,
              const access_privileges_t &privileges,
              access_flags_t             flags)
      : SmPtr(new ACL_entry_d(acl, entry_name, level, privileges, flags))
    { }

};


//---------------------------------------------------------------------

#endif  // __LNFC_ACL_H

// EOF //
