//
//  database.cpp  --  part of LNFC library
//
//  Implementation of Database class (except document search routines)
//
//  Written by Evseev at Jun 2001
//

#include "database.h"

#include "_lnfc.h"

#include <nsfdb.h>
#include <nsferr.h>    // ERR_HDL_NULL, used by RETURN_IF_NULL_HANDLE()
#include <nsfnote.h>   // NOTE_CLASS_xxx
#include <oserr.h>     // ERR_EXISTS
#include <osfile.h>    // OSPathNetParse, OSPathNetConstruct


__LNFC_NAMESPACE_BEGIN__


Database_d::Database_d(notes_handle_t active_handle,
                       const Database_path &dbpath_arg)
  : Notes_object_d(active_handle),
    dbpath(dbpath_arg)
    //dbpath(getpath(active_handle))
{
}


Database_d::Database_d(
	const Database_path &dbpath_arg,
	bool must_be_immediately_opened)
    : dbpath(dbpath_arg)
{
	if (must_be_immediately_opened)
		open();
}


int Database_d::open()
{
	close();
	RETURN_IF_ERROR();
    
	DBHANDLE htmp;
	RETURN_IF_API_ERROR(
		NSFDbOpen(dbpath->plain(), &htmp));
    
	handle = (notes_handle_t)htmp;
	return 0;
}


int Database_d::close(bool)
{
	int i = last_error;
	if (handle) {
		// don't check last_error; attempt to close anyway.
		i = status2int(NSFDbClose((DBHANDLE)handle)); // !!! throw if error
		if (ok(i)) {
			handle = 0; // clear handle only if closing was successful.
		} else if (ok()) {  // overwrite previous last_error only it was zero.
			reset(i);
		}
	}
	if (!handle) reset();    // forget last_error
	return i;                // but return it as result of close()
}


int Database_d::remove(bool force)
{
	RETURN_IF_ERROR();

	close(force);
	// we don't use RETURN_IF_ERROR() here because it ignores 'force'
	if (bad() && !force)
		return last_error;

	// we don't use RETURN_IF_API_ERROR() immediately
	// because failure is handled by calling NSFDbMarkForDelete().
	STATUS status = NSFDbDelete(dbpath->plain());
	if (status == NOERROR)
		return 0;

	if (NSFDbMarkForDelete(dbpath->plain()) != NOERROR)
		RETURN_API_ERROR(status);  // ..and only now throw saved status
	return 0;
}


int Database_d::create(bool force)
{
	close();
	RETURN_IF_ERROR();
	RETURN_IF_API_ERROR(NSFDbCreate(dbpath->plain(), DBCLASS_NOTEFILE, force));
	return open();  // throw, if cannot open
}


//------------  create FROM  ------------//


int Database_d::create(
    const Database_path &srcname,
    const Database_path &destname,
    unsigned short note_class,
    unsigned long flags,
    bool force)
{
	if (exist() && !force) {
		reset(status2int(ERR_EXISTS));
		return last_error;
	}

	close();
	RETURN_IF_ERROR();

	DBHANDLE htmp;
	RETURN_IF_API_ERROR(
		NSFDbCreateAndCopy(
			srcname->plain(),    // from filename
			destname->plain(),   // to filename
			note_class,          // NoteClass: what to copy
			0,                   // limit (0 = no limit)
			flags,               // flags
			&htmp                // result handle addr
    )   );
    handle = (notes_handle_t)htmp;
    return 0;
}


int Database_d::create(const Database_path &template_filename,
	bool force, note_class_t nclass, dbcopy_t flags)
{
	return create(template_filename, dbpath, nclass, flags, force);
}

int Database_d::create_copy_from(const Database &db,
	bool force, note_class_t nclass, dbcopy_t flags)
{
	return create(db->dbpath, dbpath, nclass, flags, force);
}

int Database_d::create_replica_from(const Database &db,
	bool force, note_class_t nclass, dbcopy_t flags)
{
	return create(db->dbpath, dbpath, nclass, DBCOPY_REPLICA | flags, force);
}


//------------  Create TO  ------------//


int Database_d::create_to(const Database_path &target_dbpath,
	bool force, note_class_t nclass, dbcopy_t flags)
{
	RETURN_IF_ERROR();

	if (!force)
		if (Database(target_dbpath)->exist())
			return status2int(ERR_EXISTS);

	DBHANDLE htmp;
	__RETURN_IF_API_ERROR(
		NSFDbCreateAndCopy(
			dbpath->plain(),        // from filename
			target_dbpath->plain(),	// to filename
			nclass,                 // NoteClass: what to copy
			0,                      // limit (0 = no limit)
			flags,
			&htmp));
    
	__RETURN_IF_API_ERROR(
		NSFDbClose(htmp));

	// ??? todo: return Database
	return 0;
}


int Database_d::create_replica_to(const Database_path &target_dbpath,
	bool force, note_class_t nclass, dbcopy_t flags)
{
	return create_to(target_dbpath, force, nclass,
		(dbcopy_t)(flags | Dbcopy_replica));
}


int Database_d::create_to(const Notes_name &server_name,
	bool force, note_class_t nclass, dbcopy_t flags)
{
	// this routine does not set 'last_error'
	// because operates on the another database

	RETURN_IF_ERROR();

	char target_dbpath[MAXPATH];
	char filename[MAXPATH];
 
	__RETURN_IF_API_ERROR(
		OSPathNetParse(dbpath->plain(), NULL, NULL, filename));
    
	__RETURN_IF_API_ERROR(
		OSPathNetConstruct(NULL, server_name->plain(), filename,
			target_dbpath));

	return create_to(Database_path(target_dbpath), force, nclass, flags);
}


int Database_d::create_replica_to(const Notes_name &server_name,
	bool force, note_class_t nclass, dbcopy_t flags)
{
	return create_to(server_name, force, nclass,
		(dbcopy_t)(flags | Dbcopy_replica));
}


int Database_d::get_title(char *title)
{
	RETURN_IF_ERROR();
	RETURN_IF_NULL_HANDLE();

	char buffer[NSF_INFO_SIZE];
    
	RETURN_IF_API_ERROR(NSFDbInfoGet((DBHANDLE)handle, buffer));
    
	NSFDbInfoParse(buffer, INFOPARSE_TITLE, title, NSF_INFO_SIZE-1);

	return 0;
}


Notes_string Database_d::get_title()
{
	char title[NSF_INFO_SIZE];
	char *p = title;
	if (get_title(title) < 0)
        	p = "(unknown title)";
	return Notes_string(p, Skip_lmbcs_conversion);
}


//   Not required now..
// int Database_d::replicate_with(const Database &db)
// {
//	return 0;
// }


Database Database::none(notes_status_t status)
{
	Database db(Notes_string("", Skip_lmbcs_conversion), Dont_open);
	db->reset(status2int(status));
	return db;
}


__LNFC_NAMESPACE_END__

// EOF //
