//
//  database.h  -- part of LNFC library
//
//  Declaration of class Database, Database_d
//
//  Written by Evseev at Jul 2001
//

#ifndef __LNFC_DATABASE_H
#define __LNFC_DATABASE_H


#ifndef __LNFC_NOBJECT_H
#include "nobject.h"
#endif

#ifndef __LNFC_DBPATH_H   
#include "dbpath.h"   //also includes nname.h, nstring.h, lnfcdefs.h, smartptr.h, cstring.h
#endif


__LNFC_NAMESPACE_BEGIN__


// forward references //
class Database;

class Document;
class Formula;


// ordinary datatypes //
typedef unsigned long  noteid_t;    // NOTEID (in fact, 16-bit?)

// NOTE_CLASS: taken from nsfnote.h
typedef enum {
	Document_note	= 0x0001,	/* document note */
	Data_note	= Document_note,/* old name for document note */
	Info_note	= 0x0002,	/* notefile info (help-about) note */
	Form_note	= 0x0004,	/* form note */
	View_note	= 0x0008,	/* view note */
	Icon_note	= 0x0010,	/* icon note */
	Design_note	= 0x0020,	/* design note collection */
	ACL_note	= 0x0040,	/* acl note */
	Help_index_note	= 0x0080,	/* Notes product help index note */
	Help_note	= 0x0100,	/* designer's help note */
	Filter_note	= 0x0200,	/* filter note */
	Field_note	= 0x0400,	/* field note */
	Replformula_note= 0x0800,	/* replication formula */
	Private_note	= 0x1000,	/* Private design note, use $PrivateDesign view to locate/classify */
	Default_note	= 0x8000,	/* MODIFIER - default version of each */
	Notify_deletion	= Default_note,	/* see SEARCH_NOTIFYDELETIONS */
	All_notes	= 0x7fff,	/* all note types */
	Allnondata_notes= 0x7ffe,	/* all non-data notes */
	No_notes	= 0x0000,	/* no notes */
	Single_instance_notes = Design_note | ACL_note | Info_note | Icon_note | Help_index_note
} note_class_t;

// DBCOPY_XXX: taken from nsfdb.h
typedef enum {
	Dbcopy_default		= 0,
	Dbcopy_replica		= 0x0001,
	Dbcopy_subclass_template= 0x0002,
	Dbcopy_dbinfo2		= 0x0004,	
	Dbcopy_special_objects	= 0x0008,
	Dbcopy_no_acl		= 0x0010,
	Dbcopy_no_fulltext	= 0x0020,
	Dbcopy_encrypt_simple	= 0x0040,
	Dbcopy_encrypt_medium	= 0x0080,
	Dbcopy_encrypt_strong	= 0x0100,
	Dbcopy_keep_note_modtime= 0x0200
} dbcopy_t;

// readable aliases for values of bool args //
const bool Dont_open = false;
const bool Force = true;
const bool Dont_force = false;
const bool Dont_open_docs = false;


//---------------------------------------------------------------------


class Database_d : public Notes_object_d
{
    protected:
	friend Database;

	// args are in reverse order for prevention ambiguity between members
	Database_d(notes_handle_t active_handle, const Database_path &dbpath);
	Database_d(const Database_path &dbpath, bool open_now = true);
	virtual ~Database_d() { cleanup(); }

    protected:
	Database_path dbpath;

    public:
	const Database_path &getpath() const
		{ return dbpath; }
	static Database_path getpath(notes_handle_t dbhandle);

    public:
	int create(bool force = false);
	int create(const Database_path &template_filename,   bool force = false, note_class_t nclass = All_notes, dbcopy_t flags = Dbcopy_default);
	int create_replica_from(const Database &db,          bool force = false, note_class_t nclass = All_notes, dbcopy_t flags = Dbcopy_default);
	int create_copy_from(   const Database &db,          bool force = false, note_class_t nclass = All_notes, dbcopy_t flags = Dbcopy_default);

	int create_to        (const Database_path &target_dbpath, bool force, note_class_t nclass, dbcopy_t flags = Dbcopy_default);
	int create_replica_to(const Database_path &target_dbpath, bool force, note_class_t nclass, dbcopy_t flags = Dbcopy_default);
	int create_to        (const Notes_name    &server_name,   bool force = false, note_class_t nclass = All_notes, dbcopy_t flags = Dbcopy_default);
	int create_replica_to(const Notes_name    &server_name,   bool force = false, note_class_t nclass = All_notes, dbcopy_t flags = Dbcopy_default);

	// int replicate_with(const Database &db);
	// bool is_replica_of(const Database &db);

	virtual int open();
	virtual int close(bool force = false);
	virtual int remove(bool force = false);
	
	bool is_remote() { return dbpath->is_remote(); }
	bool is_local() { return dbpath->is_local(); }

	Notes_string get_title();

    protected:
	int get_title(char *title);
	int create(const Database_path &srcname, const Database_path &destname,
		unsigned short note_class, unsigned long flags,
		bool force = false);

    public:
	int foreach_matching(const Formula &search_formula,
		int (*action)(Document &doc, void *args),
		void *args = 0,
		bool must_be_opened_immediately = true,
		note_class_t nclass = Data_note);
	Document first_matching(const Formula &search_formula,
		bool must_be_opened_immediately = true,
		note_class_t nclass = Data_note);
	// ... then check doc->exist()

    protected:
	int first_matching(const Formula &search_formula, noteid_t& noteid,
		note_class_t nclass = Data_note);
};


//---------------------------------------------------------------------


class Database : public Smart_pointer<Database_d>
{
    public:
	// args are in reverse order for prevention ambiguity between members
	Database(notes_handle_t active_handle, const Database_path &dbpath) :
		SmPtr(new Database_d(active_handle, dbpath))
		{ }

	Database(const Database_path &dbpath, bool open_now = true) :
		SmPtr(new Database_d(dbpath, open_now))
		{ }

	//* Database(const Database &db)
	//*   : SmPtr(&*db)
	//* { }

    protected:
	friend Database_d;

	// ...this constructor is called from Database_d::foreach_matching().
	Database(Database_d *db_addr) : SmPtr(db_addr) {}

    public:
	operator const Notes_string() { return (*this)->dbpath; }
	static Database none(notes_status_t status = 0);
};


//---------------------------------------------------------------------

__LNFC_NAMESPACE_END__

#endif

// EOF //
