//
//  directory.cpp  --  part of LNFC library
//
//  Implementation of class Directory, Directory_d
//
//  Written by Evseev at Jul 2001
//

#include "directory.h"
#include "document.h"
#include "field.h"

#include "_fileops.h"
#include <dir.h>        // mkdir(), rmdir(), searchpath()
#include <stdlib.h>     // system()
#include <strstream.h>  // strstream

#include "_lnfc.h"

#include <nsfdb.h>      // NSFDbOpen/Close/ModeGet/PathGet
#include <nsfsearc.h>   // NSFSearch(), SEARCH_MATCH
#include <nsferr.h>
#include <oserr.h>
#include <osfile.h>     // OSPathNetConstruct, FILE_xxx

#include "itemtbl.c"


__LNFC_NAMESPACE__


static Database_path users_directory(Notes_string("Users", Skip_lmbcs_conversion));

static Notes_string  lastname("Lastname", Skip_lmbcs_conversion);


#define RETURN_IFNOT_LOCAL() \
        do { if (is_remote()) RETURN_API_ERROR(ERR_NOT_LOCAL); } while(0)


#define CALLBACK_RESET(__lnfc_routine_call) \
        do { \
            int i = __lnfc_routine_call; \
            if (i < 0) \
                a->rootdir_addr->reset(i); \
        } while(0)


#define CALLBACK_RETURN_API_ERROR(__status) \
        do { \
            a->rootdir_addr->reset(status2int(__status)); \
            return __status; \
        } while(0)


#define CALLBACK_RETURN_IF_API_ERROR(__notes_api_call) \
        do { \
            STATUS status = __notes_api_call; \
            if (status != NOERROR) \
                CALLBACK_RETURN_API_ERROR(status); \
        } while(0)


//---------------------------------------------------------------------


#include <errno.h>

STATUS errno2status() {
    switch(errno) {
        case EACCES:  return ERR_ACCESS;
        case ENOENT:  return ERR_NOEXIST;
        case ENOEXEC: return ERR_OS_EXECERROR;
        case ENOMEM:  return ERR_MEMORY;
    } 
    return ERR_ACCESS;
}


//---------------------------------------------------------------------


bool we_are_under_NT()
{
#if !defined(_WIN32)
    return false;
#else
    static int result = -1;
    if (result < 0) {
        result = false;
        OSVERSIONINFO osinfo;
        osinfo.dwOSVersionInfoSize = sizeof(osinfo);
        if (GetVersionEx(&osinfo)) {
            if (osinfo.dwPlatformId == VER_PLATFORM_WIN32_NT) {
                result = true;
            }
        }
    }
    return (bool)result;
#endif
}


//---------------------------------------------------------------------


Directory_d::Directory_d(const Database_path &dirname)
  : Database_d(dirname, Dont_open)
{
    open();
}


Directory_d::Directory_d(const Document &person_doc)
  : Database_d( users_directory + FILE_SEPARATOR_STR
        + Field(person_doc, lastname)->text(), Dont_open)
{
    open();
}


int Directory_d::open()
{
    RETURN_IF_ERROR();
 // RETURN_IFNOT_LOCAL();

    Database_d::open();
    RETURN_IF_ERROR();
    RETURN_IF_NULL_HANDLE();

    USHORT dbmode;
    STATUS status;
    if (NOERROR == (status = NSFDbModeGet((DBHANDLE)handle, &dbmode))) {
        if ((dbmode & DB_DIRECTORY) != 0)
            return 0; //All Ok
        status = ERR_NOT_DIRECTORY;
    }

    close();
    RETURN_API_ERROR(status);
}


bool Directory_d::exist()
{
    if (is_open())
        return true;
    if (dbpath->is_local())
        return is_directory(dbpath->absolute()->native().c_str());
    return false;  // cannot detect remote name without opening
}


int Directory_d::create()
{
    RETURN_IFNOT_LOCAL();

    close();
    RETURN_IF_ERROR();

    if (mkdir(dbpath->absolute()->native().c_str()) < 0)
        RETURN_API_ERROR(errno2status());

    return open();
}


int Directory_d::remove(bool force)
{
    RETURN_IFNOT_LOCAL();

    close(force);
    if (!force)
        RETURN_IF_ERROR();

    Database_path fullpath = dbpath->absolute();
    
    if (rmdir(fullpath->native().c_str()) < 0)
        RETURN_API_ERROR(errno2status());

    return 0;
}


int Directory_d::open_file_access()
{
    RETURN_IF_ERROR();
    RETURN_IFNOT_LOCAL();

#if defined(_WIN32)

    if (we_are_under_NT()) {

        if (!searchpath("xcacls.exe"))
            RETURN_API_ERROR(ERR_OS_EXECERRORNF);

        char username[MAXPATH];
        DWORD ulen = MAXPATH;

        if (!GetUserName(username, &ulen))
            RETURN_API_ERROR(ERR_ACCESS);

        strstream st;
        st << "xcacls.exe " << dbpath->absolute()->native().c_str()
           << " /T /E /Y /G " << username << ":F";
        if (system(st.str()) < 0)
            RETURN_API_ERROR(errno2status());
    }

#else

#   error Directory_d::open_file_access is host-dependent and must be implemented now

#endif
    
    return 0;
}


int Directory_d::protect_file_access()
{
    RETURN_IF_ERROR();
    RETURN_IFNOT_LOCAL();
    
#if defined(_WIN32)

    if (we_are_under_NT()) {

        if (!searchpath("setperms.bat"))
            RETURN_API_ERROR(ERR_OS_EXECERRORNF);

        char drivestr[MAXDRIVE], dirstr[MAXDIR], namestr[MAXFILE], extstr[MAXEXT];
        int i = fnsplit(dbpath->absolute()->native().c_str(),
            drivestr, dirstr, namestr, extstr);

        strstream st;
        if (i & DRIVE)
            st << drivestr << " && ";
        if (i & DIRECTORY)
            st << "cd " << dirstr << " && ";
        st << "setperms.bat " << namestr << extstr;
        if (system(st.str()) < 0) {
            RETURN_API_ERROR(errno2status());
        }
    }

#else

#   error Directory_d::protect_file_access is host-dependent and must be implemented now

#endif
    
    return 0;
}


//---------------------------------------------------------------------


typedef struct {
    int(*action)(Database &db, void *args);
    void *action_args;
    bool recurse;
    bool open_now;
    Database_d *rootdir_addr;
} Foreach_args;


static STATUS LNPUBLIC foreach_enumerator(
    void *args_ptr, SEARCH_MATCH *search_match, ITEM_TABLE *summary)
{
    Foreach_args *a = reinterpret_cast<Foreach_args*>(args_ptr);

    if ((search_match->SERetFlags & SE_FMATCH) == 0)
        return NOERROR;

    char filepath[MAXPATH];
    if (!get_summary_filepath(summary, filepath))
        CALLBACK_RETURN_API_ERROR(ERR_ITEM_NOT_FOUND);
    
    char fullpath[MAXPATH];
    Database_path rootpath = a->rootdir_addr->getpath();
    Notes_string port   = rootpath->port();
    Notes_string server = rootpath->server();
    CALLBACK_RETURN_IF_API_ERROR(
        OSPathNetConstruct(
              port->length() > 0 ?   port->plain() : 0,  // port (NULL, if no port)
            server->length() > 0 ? server->plain() : "", // server ("", if local)
            filepath,
            fullpath ) );                                // returned value

    if (!is_directory_summary(summary)) {
        Database db(Notes_string(fullpath, Skip_lmbcs_conversion), a->open_now);
        CALLBACK_RESET(a->action(db, a->action_args));
    } else if (a->recurse) {
        Directory curdir(Notes_string(fullpath, Skip_lmbcs_conversion));
        CALLBACK_RESET(
            curdir->foreach_database(
                a->action,
                a->action_args,
                a->open_now,
                a->recurse));
    }
    return NOERROR;
}


int Directory_d::foreach_database(
    int(*action)(Database &db, void *args),
    void *args,
    bool open_dbs_immediately,
    bool recurse)
{
    RETURN_IF_ERROR();
    RETURN_IF_NULL_HANDLE();

    Foreach_args a;
    
    a.action = action;
    a.action_args = args;
    a.open_now = open_dbs_immediately;
    a.recurse = recurse;
    a.rootdir_addr = this;

    RETURN_IF_API_ERROR(
        NSFSearch(
            (DBHANDLE)handle,
            NULLHANDLE,        // selection formula
            NULL,              // title of view in formula
            SEARCH_FILETYPE +  // search for files
            SEARCH_SUMMARY,    // return a summary buffer
            FILE_DBANY +       // find any .NS? file
            FILE_DIRS +        // find subdirectories
            FILE_NOUPDIRS,     // don't find the ".." dir
            NULL,              // starting date
            foreach_enumerator,// call for each file found
            &a,                // argument passed to dblist_enumerator()
            NULL               // returned ending date (unused)
    )   );
    
    return 0;
}


//---------------------------------------------------------------------


typedef struct {
    const char *filename;
    char       *fullpath;
    Database_d *rootdir_addr;
} Firstmatch_args;


static STATUS LNPUBLIC firstmatch_enumerator(
    void *args_ptr, SEARCH_MATCH *search_match, ITEM_TABLE *summary)
{
    Firstmatch_args *a = reinterpret_cast<Firstmatch_args*>(args_ptr);

    if ((search_match->SERetFlags & SE_FMATCH) == 0)
        return NOERROR;

    if (!get_summary_filepath(summary, a->fullpath))
        CALLBACK_RETURN_API_ERROR(ERR_ITEM_NOT_FOUND);
    
    int namelen = strlen(a->filename);
    int pathlen = strlen(a->fullpath);

    // ??? this is incorrect, but how to detect FILE_SEPARATOR
    //     on the remote side for parsing filepath
    //     and comparing with given filename

    if (namelen < pathlen)
        if (memcmp(a->filename, a->fullpath + pathlen - namelen, namelen) == 0)
            return ERR_EXISTS;

    return NOERROR;
}


int Directory_d::search_subfile(const char *filename, char *fullpath)
{
    RETURN_IF_ERROR();
    RETURN_IF_NULL_HANDLE();

    Firstmatch_args a;
    a.filename = filename;
    a.fullpath = fullpath;
    a.rootdir_addr = this;

    STATUS status = NSFSearch(
        (DBHANDLE)handle,
        NULLHANDLE,       // selection formula
        NULL,             // title of view in formula
        SEARCH_FILETYPE + // search for files
        SEARCH_SUMMARY,   // return a summary buffer
        FILE_DBANY +      // find any .NS? file
        FILE_DIRS +       // find subdirectories
        FILE_NOUPDIRS,    // don't find the ".." dir
        NULL,             // starting date
        firstmatch_enumerator,// call for each file found
        &a,               // argument passed to dblist_enumerator()
        NULL              // returned ending date (unused)
    );
    if (status != ERR_EXISTS)
        RETURN_IF_API_ERROR(status);
    
    return 0;
}
 

Database_path Directory_d::expand(const Notes_string &subpath)
{
    if (is_local())
        return dbpath + FILE_SEPARATOR_STR + subpath;
    
    char fullpath[MAXPATH];
    if (search_subfile(subpath->native().c_str(), fullpath) < 0) {
        return dbpath + FILE_SEPARATOR_STR + subpath + " : (not found)";
    } else {
        return Database_path(fullpath, dbpath->server(), dbpath->port());
    }
}


Database Directory_d::get_database(const Database_path &dbpath, bool open_now)
{
    Database_path fullpath = expand(dbpath);
    if (ok())
        return Database(fullpath, open_now);
    return Database::none(get_last_error());
}


// EOF //
