//
//  document.cpp  --  part of LNFC library
//
//  Implementation of Document class and Database search routines
//
//  Written by Evseev at Jun 2001
//

#include "document.h"  //also includes database.h, dbpath.h, nstring.h, etc.
#include "formula.h"

#include "_lnfc.h"

#include <nsferr.h>
#include <nsfnote.h>
#include <nsfdb.h>     // NSFDbGenerateOID()

// used by Database_d search routines
#include <nsfsearc.h>  // NSFSearch()
#include <nsferr.h>    // ERR_DB_NOT_OPEN, ERR_NO_MORE_MATCHES


__LNFC_NAMESPACE_BEGIN__


Document_d::Document_d(
	const Database &db_arg,
	noteid_t noteid_arg,
	bool must_be_opened_immediately)
:
	db(db_arg),
	noteid(noteid_arg)
{
	dont_save();  // saving-on-close will be enabled by any changed field
	if (must_be_opened_immediately)
		open();
}


Document_d::Document_d(const Database &db_arg)
	: db(db_arg),
	  noteid(0)
{
	create();
}


Document_d::Document_d(notes_handle_t active_handle, const Database &db_arg)
	: Notes_object_d(active_handle),
	  db(db_arg),
	  noteid(0)
{
}


int Document_d::open(open_flags_t open_flags)
{
	RETURN_IF_DB_ISNOT_READY();

	close();
	RETURN_IF_ERROR();

	if (!noteid)
		RETURN_API_ERROR(ERR_NOTE_ID);

	NOTEHANDLE htmp;

	RETURN_IF_API_ERROR(NSFNoteOpen(
		(DBHANDLE)db->get_handle(), (NOTEID)noteid, open_flags, &htmp));
    
	handle = (notes_handle_t)htmp;
	return 0;
}


int Document_d::create()
{
	RETURN_IF_DB_ISNOT_READY();

	close();
	RETURN_IF_ERROR();

	NOTEHANDLE htmp;
	RETURN_IF_API_ERROR(NSFNoteCreate((DBHANDLE)db->get_handle(), &htmp));

	handle = (notes_handle_t)htmp;
	return 0;
}


int Document_d::close(bool)
{
	int i = last_error;
	if (handle) {
		// don't check last_error; attempt to close anyway.
		i = status2int(NSFNoteClose((NOTEHANDLE)handle)); // !!! throw if error
		if (i == NOERROR) {
			handle = 0;  // clear handle only if closing was successful.
		} else if (ok()) {   // overwrite previous last_error only it was zero.
			reset(i);
		}
	}
	if (!handle) reset();    // forget last_error
	return i;                // but return them as result of close()
}


int Document_d::save(bool force)
{
	update_flags_t update_flags = force ? UPDATE_FORCE : 0;
	return save(update_flags);
}


int Document_d::save(update_flags_t update_flags)
{
	RETURN_IF_ERROR();

	RETURN_IF_DB_ISNOT_READY();

	if (!is_open())
		RETURN_API_ERROR(ERR_INVALID_ITEMUNK);

	RETURN_IF_API_ERROR(
		NSFNoteUpdateExtended((NOTEHANDLE)handle, update_flags));
	return 0;
}


int Document_d::remove(bool force)
{
	update_flags_t update_flags = force ? UPDATE_FORCE : 0;
	return remove(update_flags);
}


int Document_d::remove(update_flags_t update_flags)
{
	RETURN_IF_ERROR();

	RETURN_IF_DB_ISNOT_READY();

	bool force = (update_flags & UPDATE_FORCE) != 0;
    
	close(force);
	// we don't use RETURN_IF_ERROR() here because it ignores 'force'
	if (bad() && !force)
		return last_error;

	RETURN_IF_API_ERROR(
		NSFNoteDelete((DBHANDLE)db->get_handle(), noteid, update_flags));

	return 0;
}


Document Document::none(notes_status_t status)
{
	Database db = Database::none();
	Document doc(db, (noteid_t)0, Dont_open);
	doc->reset(status2int(status));
	return doc;
}


int Document_d::is_correct()
{
	RETURN_IF_ERROR();
	RETURN_IF_DB_ISNOT_READY();
	RETURN_IF_NULL_HANDLE();
	RETURN_IF_API_ERROR(
		NSFNoteCheck((NOTEHANDLE)handle));
	return 0;
}


int Document_d::compute_with_form() {
	return compute_with_form(Document::none(0));
}


int Document_d::compute_with_form(const Document &form)
{
	RETURN_IF_ERROR();
	RETURN_IF_DB_ISNOT_READY();
	RETURN_IF_NULL_HANDLE();
	RETURN_IF_BAD(form);

	RETURN_IF_API_ERROR(
		NSFNoteComputeWithForm(
			(NOTEHANDLE)handle,
			(NOTEHANDLE)form->get_handle(),
			0,   // may be CWF_CONTINUE_ON_ERROR
            		0,   // error handler
            		this));

	must_be_saved();
    
	return 0;
}


int Document_d::create_copy(notes_handle_t &created_handle)
{
	created_handle = 0;
	RETURN_IF_ERROR();
	RETURN_IF_NULL_HANDLE();
    
	NOTEHANDLE new_note;
	OID new_oid;
	DBHANDLE dbhandle = (DBHANDLE)db->get_handle();

	RETURN_IF_API_ERROR(
		NSFNoteCopy((NOTEHANDLE)handle, &new_note));

	STATUS status = NSFDbGenerateOID(dbhandle, &new_oid);
	if (status != NOERROR) {
		NSFNoteClose(new_note);
		RETURN_API_ERROR(status);
	}

	NSFNoteSetInfo(new_note, _NOTE_OID, &new_oid);
	NSFNoteSetInfo(new_note, _NOTE_ID, NULL);
	NSFNoteSetInfo(new_note, _NOTE_DB, &dbhandle);

	created_handle = (notes_handle_t)new_note;
	return 0;
}


Document Document_d::create_copy()
{
	notes_handle_t created_handle = 0;
	int i = create_copy(created_handle);
	if (i < 0) {
		return Document::none(i);
	} else {
		Document doc(created_handle, db);
		doc->must_be_saved();
		return doc;
	}
}


#if 0

Document Document_d::create_encrypted_copy();

int Document_d::decrypt(Encryption_key &key, decryption_flags_t flags = 0);

#endif


int Document_d::sign(bool skip_unsigned)
{
	RETURN_IF_ERROR();
	RETURN_IF_NULL_HANDLE();

	if (skip_unsigned && !is_signed())
		return 0;

	RETURN_IF_API_ERROR(NSFNoteSign((NOTEHANDLE)handle));
	must_be_saved();
	return 0;
}

int Document_d::unsign()
{
	RETURN_IF_ERROR();
	RETURN_IF_NULL_HANDLE();
	RETURN_IF_API_ERROR(NSFNoteUnsign((NOTEHANDLE)handle));
	must_be_saved();
	return 0;
}


int Document_d::check_signed()
{
	RETURN_IF_ERROR();
	RETURN_IF_NULL_HANDLE();
	STATUS status = NSFNoteVerifySignature((NOTEHANDLE)handle,
		NULL, NULL, NULL, NULL);

	if (status == ERR_NOTE_NOT_SIGNED)
		return 0; //unsigned

	RETURN_IF_API_ERROR(status);
	return 1; //signed
}


bool Document_d::is_signed()
{
	return check_signed() == 1;
}


//---------------------------------------------------------------------
//  search documents in database
//


typedef struct {
	Database *db_addr;
	bool open_mode;
	int (*user_action)(Document &doc, void *args);
	void *user_args;
} Search_args;
    

static STATUS LNPUBLIC dbsearch_foreach_enumerator(
	void *args, SEARCH_MATCH *search_match, ITEM_TABLE* /* summary_buffer */ )
{
	int i; 
	Search_args *a = reinterpret_cast<Search_args*>(args);

	if ((search_match->SERetFlags & SE_FMATCH) == 0)
		return NOERROR;

	Document doc(*a->db_addr, search_match->ID.NoteID, a->open_mode);
	return (STATUS)a->user_action(doc, a->user_args);
}


int Database_d::foreach_matching(
	const Formula &search_formula,
	int (*action)(Document &doc, void *args),
	void *args,
	bool must_be_opened_immediately,
	note_class_t nclass)
{
	RETURN_IF_ERROR();
	RETURN_IF_NULL_HANDLE();
	RETURN_IF_BAD(search_formula);

	Database db(this);

	Search_args a;
	a.db_addr = &db;
	a.open_mode = must_be_opened_immediately;
	a.user_action = action;
	a.user_args = args;

	RETURN_IF_API_ERROR(
	    NSFSearch (
		(DBHANDLE)handle,    // database handle
		(FORMULAHANDLE)search_formula->get_handle(),  // selection formula
		NULL,                // title of view in selection formula ???
		0,                   // search flags
		nclass,              // note class to find, default = NOTE_CLASS_DATA
		NULL,                // starting date (unused)
		dbsearch_foreach_enumerator, // routine to call for each note found
		&a,                  // argument passed to enumerator routine
		NULL));              // returned ending date (unused)

	return 0;
}


static STATUS LNPUBLIC dbsearch_first_enumerator(
	void *args, SEARCH_MATCH *search_match, ITEM_TABLE* /* summary_buffer */ )
{
	int i; 
	noteid_t *noteid_addr = reinterpret_cast<noteid_t*>(args);

	if ((search_match->SERetFlags & SE_FMATCH) == 0)
		return NOERROR;

	*noteid_addr = search_match->ID.NoteID;
	return ERR_NO_MORE_MATCHES;
}


int Database_d::first_matching(
    const Formula &search_formula, noteid_t& noteid, note_class_t nclass)
{
	RETURN_IF_ERROR();
	RETURN_IF_NULL_HANDLE();
	RETURN_IF_BAD(search_formula);

	noteid = 0;

	STATUS status = NSFSearch (
        	(DBHANDLE)handle,    // database handle
		(FORMULAHANDLE)search_formula->get_handle(),  // selection formula
		NULL,                // title of view in selection formula ???
		0,                   // search flags
		nclass,              // note class to find, default = NOTE_CLASS_DATA
		NULL,                // starting date (unused)
		dbsearch_first_enumerator,  // routine to call for each note found
		&noteid,             // argument passed to enumerator routine
		NULL);               // returned ending date (unused)

	if (noteid != 0)
		return 0;

	if (status == NOERROR)
		RETURN_API_ERROR(ERR_NO_MORE_MATCHES);
    
	RETURN_API_ERROR(status);
}

      
Document Database_d::first_matching(
	const Formula &search_formula,
	bool must_be_opened_immediately,
	note_class_t nclass)
{
	noteid_t noteid = 0;

	if (bad(first_matching(search_formula, noteid, nclass)))  // !!! throw if error
		return Document::none(ERR_GETDOCFIELD_NOTENOTFOUND);
    
	Database db(this);
	return Document(db, noteid, must_be_opened_immediately);
}


__LNFC_NAMESPACE_END__

// EOF
