//
//  field.cpp  --  part of LNFC library
//
//  Implementation of Field class
//

#include "field.h"

#include "_lnfc.h"

#include <nsfnote.h>
#include <nsferr.h>
#include <misc.h>    // TIMEDATE_MINIMUM


// this macro copies last_error from document, if document is bad
#define IF_ISNOT_READY(_action) \
        do { \
            if( bad() || \
                !doc->is_open() || \
                ( doc->bad() && \
                  ( reset(doc->get_last_error()), true ) \
              ) ) \
            { _action; } \
        } while(0)


#define RETURN_IF_ISNOT_READY() \
        do { \
            if (bad()) \
                return last_error; \
            if (!doc->is_open()) \
                RETURN_API_ERROR(ERR_DB_NOT_OPEN); \
            RETURN_IF_BAD(doc); \
        } while(0)

//
//  To do: for each of the following data types,
//  the data pointed to by item_value must be in Notes canonical format. 
//
//  Program must convert the item value to Notes Canonical Format
//  using ODSWriteMemory() before calling NSFItemAppend() if:
//    1) it runs on a non-Intel architecture machine, and 
//    2) item_type is one of the following data types: 
//
#ifdef _M_IX86      /* this is Borland C++ macro */
# define RETURN_IF_UNSUPPORTED_TYPE(t)
#else
# define RETURN_IF_UNSUPPORTED_TYPE(t) \
    do {                              \
        switch(t) {                   \
        case TYPE_FORMULA:            \
        case TYPE_COMPOSITE:          \
        case TYPE_COLLATION:          \
        case TYPE_OBJECT:             \
        case TYPE_VIEW_FORMAT:        \
        case TYPE_ICON:               \
        case TYPE_SIGNATURE:          \
        case TYPE_SEAL:               \
        case TYPE_SEALDATA:           \
        case TYPE_SEAL_LIST:          \
        case TYPE_WORKSHEET_DATA:     \
        case TYPE_USERDATA:           \
            RETURN_API_ERROR(ERR_UNK_TYPE); \
        } /*switch*/                  \
    } while(0)
#endif  // !defined(_M_IX86)


#define DOC_IS_CHANGED()  doc->must_be_saved()


__LNFC_NAMESPACE_BEGIN__


const Field_flags Field::TEXT    = mkdword( 0, TYPE_TEXT );
const Field_flags Field::LIST    = mkdword( 0, TYPE_TEXT_LIST );
const Field_flags Field::NAMES   = mkdword( ITEM_NAMES, TYPE_TEXT_LIST );
const Field_flags Field::AUTHORS = mkdword( ITEM_READWRITERS | ITEM_NAMES, TYPE_TEXT_LIST );


const bool Skips_lmbcs_conversion = true;


Field_d::Field_d(const Document &doc_arg, const Notes_string &field_name_arg)
  : doc(doc_arg),
    field_name(field_name_arg)
{
}


int Field_d::assign(const Notes_string &new_value)
{
    RETURN_IF_ISNOT_READY();

    char *data_addr = new_value->plain();
    WORD  data_len  = new_value->length();

    RETURN_IF_API_ERROR(
        NSFItemSetText(
            (NOTEHANDLE)doc->get_handle(),
            field_name->plain(),
            data_addr,
            data_len));

    DOC_IS_CHANGED();

    return 0;
}


int Field_d::assign(double new_value)
{
    RETURN_IF_ISNOT_READY();

    RETURN_IF_API_ERROR(
        NSFItemSetNumber(
            (NOTEHANDLE)doc->get_handle(),
            field_name->plain(),
            &new_value));

    DOC_IS_CHANGED();

    return 0;
}


#pragma argsused
int Field_d::assign(const Timedate &timedate)
{
    return 0;     // ???
}


int Field_d::create(Field_flags field_flags, const Notes_string &new_value)
{
    RETURN_IF_ISNOT_READY();

    RETURN_IF_UNSUPPORTED_TYPE(field_flags.get_item_type());

    RETURN_IF_API_ERROR(
        NSFItemAppend(
            (NOTEHANDLE)doc->get_handle(),
            field_flags.get_item_flags(),
            field_name->plain(),
            field_name->length(),
            field_flags.get_item_type(),
            new_value->plain(),
            new_value->length() ));

    DOC_IS_CHANGED();

    return 0;
}


int Field_d::create(Field_flags field_flags, double new_value)
{
    RETURN_IF_ISNOT_READY();

    RETURN_IF_UNSUPPORTED_TYPE(field_flags.get_item_type());

    RETURN_IF_API_ERROR(
        NSFItemAppend(
            (NOTEHANDLE)doc->get_handle(),
            field_flags.get_item_flags(),
            field_name->plain(),
            field_name->length(),
            TYPE_NUMBER, // ignores field_flags.get_item_type()
            &new_value,
            sizeof(new_value) ));

    DOC_IS_CHANGED();

    return 0;
}


#pragma argsused
int Field_d::create(Field_flags field_flags, const Timedate &timedate)
{
    return 0;   // ???
}


bool Field_d::exist()
{
    IF_ISNOT_READY(return false);

    return (TRUE == NSFItemIsPresent(
        (NOTEHANDLE)doc->get_handle(),
        field_name->plain(),
        field_name->length()));
}


int Field_d::remove()
{
    RETURN_IF_ISNOT_READY();

    RETURN_IF_API_ERROR(
        NSFItemDelete(
            (NOTEHANDLE)doc->get_handle(),
            field_name->plain(),
            field_name->length()
        )
    );

    return 0;
}


int Field_d::add_line(const Notes_string &line, bool allow_duplicates)
{
    RETURN_IF_ISNOT_READY();
    
    RETURN_IF_API_ERROR(
        NSFItemAppendTextList(
            (NOTEHANDLE)doc->get_handle(),
            field_name->plain(),
            line->plain(),
            line->length(),
            allow_duplicates));

    DOC_IS_CHANGED();

    return 0;
}


static const size_t linepos(
    const Notes_string &where, const Notes_string &what, size_t startpos = 0)
{
    size_t n;
    while ((n = where->find_first_of(*what, startpos)) != NPOS) {
        if ( ( (n == 0) ||
               ((*where)[n - 1] == '\0') ) &&
             ( (n + what->length() > where->length()) ||
               ((*where)[n + what->length()] == '\0')
           ) )
        {
            //start = n + where->length();  // incorrect in conjunction with erase()
            return n;
        }
        startpos = n;
    }
    return NPOS;
}


int Field_d::remove_line(const Notes_string &line)
{
    RETURN_IF_ISNOT_READY();
    
    Notes_string s = text('\0');
    if (s->is_empty())
        RETURN_API_ERROR(ERR_ITEM_LENGTH);  // catch possible error in text()

    size_t pos = 0;
    bool is_changed = false;

    while ((pos = linepos(s, line, pos)) != NPOS) {
        is_changed = true;
        s->erase(pos, line->length());
    }

    if (is_changed)
        assign(s);   // DOC_IS_CHANGED() will be called in assign()

    return 0;
}


bool Field_d::contains_line(const Notes_string &line)
{
    IF_ISNOT_READY(return false);

    Notes_string s = text('\0');
    if (s->is_empty()) return false;   // catch possible error in text()

    return linepos(s, line) != NPOS;
}


int Field_d::lines_count()
{
    IF_ISNOT_READY(return 0);

    return NSFItemGetTextListEntries(
        (NOTEHANDLE)doc->get_handle(),
        field_name->plain()
    );
}


//---------------------------------------------------------------------
//
//   Field_d::line() is very composite routine
//   because it implements common allocation scheme of temporary buffer.
//   This scheme is used for 3 various NotesAPI calls:
//     - get selected line of multi-line text field,
//     - get field contents in raw binary form ('\0' separated),
//     - get field contents in readable form
//       (string that contains lines of field separated by given char).
//   In all three cases we cannot get actual data length
//   without already existing buffer for save
//   (through fool programming style of Notes creators).
//   So, length of buffer (with following sub-routine calling)
//   is increased in loop until it will be enough to store requested data.
//
Notes_string Field_d::line(int linenum, char list_separator)
{
    IF_ISNOT_READY(return "");

    char startbuf[1024];  // is 1k enough?
    size_t bufsize = sizeof(startbuf);
    char *buf = startbuf;

    do {
        WORD actual_len;
        NOTEHANDLE htmp = (NOTEHANDLE)doc->get_handle();
        
        if (linenum >= 0) {          // return selected line of item
            actual_len = NSFItemGetTextListEntry(
                htmp, field_name->plain(), linenum, buf, bufsize-1);
        } else if (list_separator) { // return all contents of item in text form
            actual_len = NSFItemConvertToText(
                htmp, field_name->plain(), buf, bufsize-1, list_separator);
        } else {                  // return all contents of item in binary form
            actual_len = NSFItemGetText(
                htmp, field_name->plain(), buf, bufsize-1);
        }

        if ((size_t)actual_len < bufsize-1) {
            Notes_string s(buf, Skips_lmbcs_conversion, 0, actual_len);
            
            if (buf != startbuf)
                delete buf;
            
            return s;
        }

        if (buf != startbuf)
            delete buf;
        buf = new char[bufsize *= 2];

    } while(bufsize < max_possible_signed<size_t>() &&
            bufsize < (size_t)max_possible_unsigned<WORD>());

    return "";  // !!! throw if error
}


long Field_d::ordinal_value(long default_value)
{
    IF_ISNOT_READY(return default_value);

    return NSFItemGetLong(
           (NOTEHANDLE)doc->get_handle(),
           field_name->plain(),
           default_value);
}


double Field_d::value(double default_value)
{
    IF_ISNOT_READY(return default_value);

    double result;

    if (NSFItemGetNumber(
        (NOTEHANDLE)doc->get_handle(),
        field_name->plain(),
        &result) == NOERROR)
    {
        return result;
    }
    return default_value;
}


Timedate Field_d::timedate_value()
{
    notes_timedate_t no_time;
    TimeConstant(TIMEDATE_MINIMUM, (TIMEDATE*)&no_time);

    IF_ISNOT_READY(return no_time);

    notes_timedate_t result;

    if (NSFItemGetTime(
        (NOTEHANDLE)doc->get_handle(),
        field_name->plain(),
        (TIMEDATE*)&result) == NOERROR)
    {
        return result;
    }
    return no_time;
}


__LNFC_NAMESPACE_END__

// EOF
