//
//  nsfcopy.cpp
//
//  Written at Jan,Feb 2003 by <evseev@ucvt.ru>
//

#include <string.h>
#include <fstream.h>

#pragma hdrstop

#include "nsession.h"
#include "nstring.h"
#include "nobject.h"
#include "database.h"
#include "document.h"
#include "formula.h"

#if defined(_WIN32)
#  define cerr cout
#  define SLASH_STR "\\"
#else
#  define SLASH_STR "/"
#endif

#define C_STR(x)  (((string)x).c_str())

typedef enum { Error_output, Quiet_output, Normal_output, Verbose_output }
	Output_level;
Output_level output_level = Normal_output;

bool do_replication;

string extract_filename(const string &fpath)
{
	int i = fpath.rfind(SLASH_STR);
	if (i > 0)
		return string(fpath, i+1);
	return fpath;
}

void title(bool force = false)
{
	static bool passed;
	if (passed)
		return;
	passed = true;
	if (output_level < Normal_output && !force)
		return;
	cout << "NSFCopy 0.2, written at Feb 2003 by evseev@ucvt.ru" << endl;
}

int report_nonewline(Output_level required_level, const string s)
{
	if (required_level > output_level)
		return -1;
	title();
	cout << s;
	return s.length();
}

int report(Output_level required_level, const string s)
{
	int i = report_nonewline(required_level, s);
	if (i >= 0)
		cout << endl;
	return i;
}

int usage(const char *msg = 0)
{
        title(true);
	if (msg) {
		cerr << "Error: " << msg << endl;
	}
	cerr << "Usage: nsfcopy [options] file.nsf... destination-dir\n"
		"Where options are:\n"
		" -r   Create replicas (by default, create new copy)\n"
		" -s   Resign signed docs by current ID\n"
		" -S   Sign all docs by by current ID\n"
		" -e   Encrypt created databases by current ID (simple)\n"
		" -E   Encrypt created databases by current ID (medium)\n"
		" -!   Encrypt created databases by current ID (strong)\n"
		" -f   Overwrite existing destination databases\n"
		" -q   Quiet output\n"
		" -v   Verbose output" << endl;
	return 1;
}

int sign_doc(Document &doc, void* arg)
{
	bool *force_all = (bool*)arg;
	return doc->sign(! *force_all);
}

int main(int argc, char *argv[])
{
    //	cout << "Before init..." << endl;
	INIT_NOTES_RUNTIME();
    //	cout << "After init..." << endl;

	dbcopy_t flags = (dbcopy_t)(Dbcopy_default | Dbcopy_no_fulltext);
	bool force_overwrite = false;
	bool sign_it = false;
	bool force_sign = false;
        Formula search_all(nSTR "@All");
	int n;

	for (n = 1; n < argc; n++) {
		char *s = argv[n];
		if (s[0] != '-')
			break;
		if (s[2])
			return usage(C_STR("Unexpected characters in " + s));
		switch(s[1]) {
			case 'r': do_replication = true;	break;
			case 'q': output_level = Quiet_output;	break;
			case 'v': output_level = Verbose_output; break;
			case 'e': flags = (dbcopy_t)(flags | Dbcopy_encrypt_simple); break;
			case 'E': flags = (dbcopy_t)(flags | Dbcopy_encrypt_medium); break;
			case '!': flags = (dbcopy_t)(flags | Dbcopy_encrypt_strong); break;
			case 's': sign_it = true;		break;
			case 'S': sign_it = force_sign = true;  break;
			case 'f': force_overwrite = true;	break;
			default : return usage(C_STR("Unknown switch " + s));
		}
	}

	if (argc - n < 2)
		return usage();

	title();

	string destdir = argv[argc-1];
	if (destdir[destdir.length() - 1] != SLASH_STR[0])
		destdir += SLASH_STR;
	char *action_name[] = { "copying", "replication" };
	report(Verbose_output, C_STR("Starting " + action_name[do_replication] +
		", destination = " + destdir));

	for ( ; n < argc - 1; n++) {
		Database_path srcpath(argv[n]);
		Database srcdb(srcpath);
		if (!srcdb->exist()) {
			report(Error_output, srcdb->error_string(
				C_STR("cannot open source database: " +
				srcpath->readable())));
			continue;
		}

		Database destdb(dbPATH destdir + extract_filename(srcpath->readable()));
		if (destdb->exist()) {
			report(force_overwrite ? Normal_output : Error_output,
				C_STR("Destination database already exists: " +
				destdb->getpath()->readable()));
			if (!force_overwrite)
				continue;
		}

		report_nonewline(Normal_output, C_STR(srcdb->getpath()->readable() + "..."));
		
		if (do_replication) {
			destdb->create_replica_from(srcdb, force_overwrite, All_notes, flags);
		} else {
			destdb->create_copy_from(srcdb, force_overwrite, All_notes, flags);
		}
		if (sign_it) {
			destdb->foreach_matching(
				search_all, sign_doc, &force_sign, true, All_notes);
		}

		if (destdb->ok()) {
			report(Normal_output, " OK");
		} else {
			report(Normal_output, "");
			report(Error_output, destdb->error_string(
			        C_STR("Cannot write destination " +
				destdb->getpath()->readable())));
		}
	}

	return 0;
}

// EOF //
