//
//  smartptr.h
//
//  Basic classes for creating Java-like counter-based
//  reference-accessed auto-removable classes #;-P
//
//  CONTENTS:
//
//     Self_destroyed class: root class for data containers
//
//        Contains counter and deallocates self when counter become to 0.
//        Must be allocated on heap only!
//        Must be not created by user directly,
//        only from associated Smart_pointer<> !
//
//     Smart_pointer class template: root for data reference type
//
//        Contain link to object derived from Self_destroyed.
//        //Automatically creates associated object in default constructor.
//        Overload assignment operator for modifying counter of associated obj
//        Overload dereference operator for transparent access to assoc. object
//
//     Smart_pointer2 class template:
//
//        Is derived from Smart_pointer<Basic_data> and Smart_pointer<Ext_data>,
//        where Ext_data is derived from Basic_data.
//
//        As child of SmartPointer<Ext_data>,
//        it can be used for access to members of Ext_data.
//
//        As child of SmartPointer<Basic_data>, it can be used for passing
//        as parameter to routine that expects SmartPointer<Basic_data>,
//        for transparently access to Basic_data part of Ext_data.
//
//  Written by Ilya Evseev (ilya_evseev@chat.ru) at Jun 2001
//

#ifndef __SMARTPTR_H
#define __SMARTPTR_H


#ifndef  THROW_NONE
#define  THROW_NONE  /* throw() */
#endif

typedef unsigned long counter_t;

#define atomic_inc(value_ptr)  (++(*(value_ptr)))
#define atomic_dec(value_ptr)  (--(*(value_ptr)))


/*-------------------------------------------------------------
 *  Initially_zero class template
 */
template<class T> struct Initially_zero {
    T  value;
    Initially_zero(void)  : value(0) {}
    Initially_zero(T val) : value(val) {}
    T* operator & (void) { return  &value; }
    T  operator ++(void) { return ++value; }
    T  operator --(void) { return --value; }
       operator T (void) { return   value; }
};


/*-------------------------------------------------------------
 *  Self_destroyed class
 */
class Self_destroyed {

public:

    virtual ~Self_destroyed() {
    }

    counter_t _addref() {
        return atomic_inc(&nlocks);
    }

    counter_t _release() {
        counter_t tmp;
        if ((tmp = atomic_dec(&nlocks)) <= 0)
            delete this;
        return tmp;
    }

private:

    Initially_zero<counter_t> nlocks;
};


/*-------------------------------------------------------------
 *  Smart_pointer template
 */
template<class X> class Smart_pointer
{

protected:

    typedef Smart_pointer<X> SmPtr;
 // typedef X element_type;

 // explicit Smart_pointer()  THROW_NONE
 //   : associated_object(0)
 // { }

    explicit Smart_pointer(X *p)  THROW_NONE {
        _setup(p);
    }

    Smart_pointer(const Smart_pointer<X> &a)  THROW_NONE {
        _setup(a.associated_object);
    }

    Smart_pointer<X>& operator = (const Smart_pointer<X> &another)  THROW_NONE {
        return _assign(another.associated_object);
    }

    ~Smart_pointer()  THROW_NONE {
        if (associated_object)
            associated_object->_release();
    }
    
public:

    // ----  members  ----------------------------
    //
    X& operator*  ()  const THROW_NONE { return *associated_object;   }
    X* operator-> ()  const THROW_NONE { return associated_object;    }
    //X* get      ()  const THROW_NONE { return associated_object;    }

protected:

    void _setup(X *p)  THROW_NONE
    { 
        if ((associated_object = p) == 0) {
            // associated_object = new X();
        } else {
            // !!! throw
        }
        if (associated_object)
            associated_object->_addref();
    }

    Smart_pointer<X>& _assign(X *p)  THROW_NONE
    { 
        X *old = associated_object;
        _setup(p);
        if (old) old->_release();
        return *this;
    }

protected:

    X* associated_object;
};


/*---------------------------------------------------------------------
 *  Smart_pointer2 template class
 */
template<class RefType, class DataType> class Smart_pointer2
  : public Smart_pointer<DataType>,
    public RefType
{

protected:

  //typedef RefType::SmPtr RefSmPtr;  //not used anywhere and causes error under GCC
    typedef Smart_pointer<DataType> SmPtr;
    typedef Smart_pointer2<RefType, DataType> SmPtr2;

    explicit Smart_pointer2(DataType *p)  THROW_NONE
      : SmPtr(p),
        RefType(p)
    { }

    Smart_pointer2(SmPtr2 &a)  THROW_NONE
      : SmPtr  (a.SmPtr::associated_object),
        RefType(a.SmPtr::associated_object)
    { }

public:

    DataType& operator* () const THROW_NONE {return *SmPtr::associated_object;}
    DataType* operator->() const THROW_NONE {return  SmPtr::associated_object;}
    //* DataType* get   () const THROW_NONE {return  SmPtr::associated_object;}
};


#endif  // ifndef __SMARTPTR_H

// EOF
