//
//   view.cpp  --  part of LNFC library
//
//   Implementation of classes View, View_d
//
//   Written by Evseev at Aug 2001
//


#include "view.h"

#include "field.h"  // !!! required for lookup_by_field() only

#include "_lnfc.h"


#include <nif.h>    // COLLECTIONPOSITION
#include <niferr.h> // ERR_NO_MORE_MEMBERS
#include <nsferr.h> // ERR_HDL_NULL, used by RETURN_IF_NULL_HANDLE()
#include <osmem.h>  // OSLock/UnlockObject, OSMemFree
#include <oserr.h>  // ERR_MEMORY


#define  NOTEID_THAT_MUST_BE_SKIPPED1  0x80000004
#define  NOTEID_THAT_MUST_BE_SKIPPED2  0x80000008


__LNFC_NAMESPACE_BEGIN__


//---------------------------------------------------------------------


class View_d : public View_i
{

public:

    View_d(const Database &db, const Notes_string &view_name);

    virtual Document at(size_t index, bool open_now = true);

    virtual Document next_document(bool open_now = true);

    virtual void seek_to_first();

    virtual bool is_end_reached() {
        return last_error == status2int(ERR_NO_MORE_MEMBERS);
    }

    virtual Document lookup(const Notes_string &sortkey_value, bool open_now = true);

    virtual Document lookup_by_column(int column, const Notes_string &value, bool open_now = true);

    virtual Document lookup_by_field(const Notes_string &field_name,
                                     const Notes_string &field_value);

    virtual int update();

    virtual Database get_database() {
        return db;
    }

protected:

    int open();

    int at(COLLECTIONPOSITION &position, size_t index, noteid_t &noteid);

    int at(size_t index, noteid_t &noteid);

    int next_document(noteid_t &noteid);

    int lookup(const Notes_string &sortkey_value, noteid_t &noteid);

    int lookup_by_column(int column, const Notes_string &value, noteid_t &noteid);

    Document final_things(int retval, noteid_t &noteid, bool open_now);

private:

    Database db;

    Notes_string view_name;

    COLLECTIONPOSITION current_position;

};


//---------------------------------------------------------------------


View_d::View_d(const Database &db_arg, const Notes_string &view_name_arg)
  : //View_i(db_arg, view_name_arg),
    db(db_arg),
    view_name(view_name_arg)
{
    open();
}


int View_d::open()
{
    RETURN_IF_BAD(db);

    if (!db->is_open()) {
        db->open();
        RETURN_IF_BAD(db);
    }

    DBHANDLE dbhandle = (DBHANDLE)db->get_handle();

    NOTEID noteid;

    RETURN_IF_API_ERROR(
        NIFFindView(
            dbhandle,
            view_name->plain(),
            &noteid
        )
    );

    HCOLLECTION collection;

    RETURN_IF_API_ERROR(
        NIFOpenCollection(
            dbhandle,
            dbhandle,
            noteid,
            0,           // ??? OPEN_FLAGS - optional argument must be passed
            NULLHANDLE,  // IN hUnreadList
            &collection,
            NULL,        // OUT &hViewNote
            NULL,        // OUT &ViewUNID
            NULL,        // OUT &hCollapsedList
            NULL         // OUT &hSelectedList
        )
    );

    handle = (notes_handle_t)collection;

    seek_to_first();

    return 0;
}


int View_d::update()
{
    RETURN_IF_ERROR();

    RETURN_IF_NULL_HANDLE();

    RETURN_IF_API_ERROR(NIFUpdateCollection((HCOLLECTION)handle));

    return 0;
}


void View_d::seek_to_first()
{
    current_position.Level = 0;       // Tumbler[].length
    current_position.Tumbler[0] = 0;  // seek BEFORE first document

    if (is_end_reached())
        reset();
}


int View_d::at(COLLECTIONPOSITION &position, size_t index, noteid_t &noteid)
{
    RETURN_IF_ERROR();

    RETURN_IF_NULL_HANDLE();

    HANDLE htmp;
    WORD sigflags;
    DWORD nskipped, nreturned;

    RETURN_IF_API_ERROR(
        NIFReadEntries(
            (HCOLLECTION)handle,
            &position,
            NAVIGATE_NEXT,
            index,            // count of entries to skip before
            NAVIGATE_NEXT,
            1,                // entries to query
            READ_MASK_NOTEID,
            &htmp,            // OUT buffer to store result
            NULL,             // OUT *wBufferLength
            &nskipped,        // OUT *dwNumEntriesSkipped, may be NULL
            &nreturned,       // OUT *dwNumEntriesReturned, may be NULL
            &sigflags         // OUT *wSignalFlags
        )
    );

    // ???  if (sigflags & SIGNAL_ANY_CONFLICT)

    if (htmp) {
        // NOTEID* is more accurately here than noteid_t*
        NOTEID *p = (NOTEID*)OSLockObject(htmp);
        noteid = *p;
        OSUnlockObject(htmp);
        OSMemFree(htmp);

        if (noteid == NOTEID_THAT_MUST_BE_SKIPPED1 ||
            noteid == NOTEID_THAT_MUST_BE_SKIPPED2)
        {
            return at(position, 1, noteid);
        }
    }

    return 0;
}


int View_d::at(size_t index, noteid_t &noteid)
{
    if (is_end_reached())
        reset();  // clear error state

    COLLECTIONPOSITION position;

    position.Level = 0;        // length of Tumbler[] - 1
    position.Tumbler[0] = 1;   // position, started from 1

    return at(position, index, noteid);
}


Document View_d::at(size_t index, bool open_now)
{
    noteid_t noteid = 0;    
    return final_things(at(index, noteid), noteid, open_now);
}


int View_d::next_document(noteid_t &noteid)
{
    return at(current_position, 1, noteid);
}


Document View_d::next_document(bool open_now)
{
    noteid_t noteid = 0;
    return final_things(next_document(noteid), noteid, open_now);
}


int View_d::lookup(const Notes_string &sortkey_value, noteid_t &noteid)
{
    RETURN_IF_ERROR();

    RETURN_IF_NULL_HANDLE();

    COLLECTIONPOSITION position;
    DWORD nmatches;

    RETURN_IF_API_ERROR(
        NIFFindByName(
            (HCOLLECTION)handle,
            sortkey_value->plain(),
            FIND_EQUAL,   // ??? FIND_FLAGS : FIND_PARTIAL, etc.
            &position,
            &nmatches
        )
    );

    // ??? if (nmatches > 1) ;

    return at(position, 0, noteid);
}


Document View_d::lookup(const Notes_string &sortkey_value, bool open_now)
{
    noteid_t noteid = 0;
    return final_things(lookup(sortkey_value, noteid), noteid, open_now);
}


int View_d::lookup_by_column(int column, const Notes_string &value, noteid_t &noteid)
{
    RETURN_IF_ERROR();

    RETURN_IF_NULL_HANDLE();

    //-------------------------------------------------

    unsigned keylen = value->length();
    unsigned bufsize = sizeof(ITEM_TABLE) + column*(sizeof(ITEM) + sizeof(WORD)) + keylen;

    char *keybuf = new char[bufsize];
    char *p = keybuf;
    int i;

    if (!keybuf)
        RETURN_IF_API_ERROR(ERR_MEMORY);

    //  produce ITEM_TABLE header
    //
    ((ITEM_TABLE*)p)->Length = bufsize;
    ((ITEM_TABLE*)p)->Items  = column;
    p += sizeof(ITEM_TABLE);

    //  produce headers of items
    //
    for (i = 0; i < column-1; i++) {
        ((ITEM*)p)->NameLength = 0;
        ((ITEM*)p)->ValueLength = sizeof(WORD);
        p += sizeof(ITEM);
    }

    //  produce header of last item
    //
    ((ITEM*)p)->NameLength = 0;
    ((ITEM*)p)->ValueLength = keylen + sizeof(WORD);
    p += sizeof(ITEM);

    //  produce pairs (name, (type, value)).
    //  In fact, all these are empty except (type,value) of last item
    //
    for (i = 0; i < column; i++)
        *((WORD*)p)++ = TYPE_TEXT;
    
    memcpy(p, value->plain(), keylen);

    //-------------------------------------------------

    COLLECTIONPOSITION position;
    DWORD nmatches;

    STATUS status = NIFFindByKey(
            (HCOLLECTION)handle,
            keybuf,
            FIND_EQUAL | FIND_PARTIAL,
            &position,
            &nmatches);

    delete keybuf;

    RETURN_IF_API_ERROR(status);

    // ??? if (nmatches > 1) ;

    return at(position, 0, noteid);
}


Document View_d::lookup_by_column(int column, const Notes_string &value, bool open_now)
{
    noteid_t noteid = 0;
    return final_things(lookup_by_column(column, value, noteid), noteid, open_now);
}


Document View_d::lookup_by_field(const Notes_string &field_name,
                                 const Notes_string &field_value)
{
    // !!! must be improved !!!

    // RETURN_IF_ERROR();

    seek_to_first();

    while(1) {
        Document doc = next_document();
        if (last_error)
            break;
        Field f(doc, field_name);
        if (f->text() == field_value)
            return doc;
    }

    return Document::none(last_error);
}


Document View_d::final_things(int retval, noteid_t &noteid, bool open_now)
{
    if (retval < 0)
        return Document::none(last_error);

    if (!noteid) {
        int i = status2int(ERR_NO_MORE_MEMBERS);
        reset(i);
        return Document::none(i);
    }

    return Document(db, noteid, open_now);
}


//---------------------------------------------------------------------


View::View(const Database &db, const Notes_string &view_name)
  : SmPtr(new View_d(db, view_name))
{ }


//---------------------------------------------------------------------

__LNFC_NAMESPACE_END__

// EOF //
